@@ -7,8 +7,9 @@
 //! ```
 //! use fortify::*;
 //!
-//! // Define a borrowing type.
-//! #[derive(WithLifetime)]
+//! // Define a borrowing type. The `Lower` trait specifies that it is covariant in its first
+//! // lifetime parameter.
+//! #[derive(Lower)]
 //! struct Example<'a> {
 //!    a: &'a i32,
 //!    b: &'a mut i32,
@@ -36,10 +37,12 @@
 //! });
 //! ```
 extern crate self as fortify;
+mod lower;
+
 pub use fortify_derive::*;
+pub use lower::*;
 use std::future::Future;
-use std::marker::PhantomData;
-use std::mem::{forget, transmute, transmute_copy, ManuallyDrop, MaybeUninit};
+use std::mem::{transmute_copy, ManuallyDrop, MaybeUninit};
 use std::pin::Pin;
 use std::task::{Context, Poll, RawWaker, RawWakerVTable, Waker};
 
@@ -73,39 +76,39 @@ impl<T> Fortify<T> {
             data_drop_fn: drop_nop,
         }
     }
+}
 
+impl<'a, T: Lower<'a, Target = T> + 'a> Fortify<T> {
     /// Creates a [`Fortify`] by explicitly providing its owned data and constructing its value
     /// from that using a closure. Note that for technical reasons, the constructed value must be
-    /// wrapped in a [`FortifySource`] wrapper.
+    /// wrapped in a [`Lowered`] wrapper.
     ///
     /// # Example
     /// ```
-    /// use fortify::{Fortify, FortifySource};
+    /// use fortify::{Fortify, Lowered};
     /// let mut str = String::new();
     /// str.push_str("Hello");
-    /// let fortified: Fortify<&str> = Fortify::new_dep(str, |s| FortifySource::new(s.as_str()));
+    /// let fortified: Fortify<&str> = Fortify::new_dep(str, |s| Lowered::new(s.as_str()));
     /// assert_eq!(fortified.borrow(), &"Hello");
     /// ```
-    pub fn new_dep<'a, O: 'a, C>(owned: O, cons: C) -> Self
+    pub fn new_dep<O: 'a, C>(owned: O, cons: C) -> Self
     where
-        T: WithLifetime<'a, Target = T>,
-        C: 'a + for<'b> FnOnce(&'b mut O) -> FortifySource<'b, 'a, T>,
+        C: 'a + for<'b> FnOnce(&'b mut O) -> Lowered<'b, T>,
     {
         Self::new_box_dep(Box::new(owned), cons)
     }
 
     /// Creates a [`Fortify`] by explicitly providing its owned data (as a [`Box`]) and
     /// constructing its value from that using a closure. Note that for technical reasons, the
-    /// constructed value must be wrapped in a [`FortifySource`] wrapper.
-    pub fn new_box_dep<'a, O: 'a, C>(owned: Box<O>, cons: C) -> Self
+    /// constructed value must be wrapped in a [`Lowered`] wrapper.
+    pub fn new_box_dep<O: 'a, C>(owned: Box<O>, cons: C) -> Self
     where
-        T: WithLifetime<'a, Target = T>,
-        C: 'a + for<'b> FnOnce(&'b mut O) -> FortifySource<'b, 'a, T>,
+        C: 'a + for<'b> FnOnce(&'b mut O) -> Lowered<'b, T>,
     {
         let owned = Box::into_raw(owned);
         let value = cons(unsafe { &mut *owned });
         Self {
-            value: ManuallyDrop::new(value.into_inner()),
+            value: ManuallyDrop::new(Lowered::unwrap(value)),
             data_raw: owned as *mut (),
             data_drop_fn: drop_box_from_raw::<O>,
         }
@@ -124,46 +127,40 @@ impl<T> Fortify<T> {
     ///
     /// # Example
     /// ```
-    /// use fortify::Fortify;
+    /// use fortify::{Fortify, Lowered};
     /// let external = 1;
     /// let mut fortified: Fortify<(&i32, &i32)> = Fortify::new_async(|y| async {
     ///     let internal = 2;
-    ///     y.yield_((&external, &internal)).await;
+    ///     y.yield_(Lowered::new((&external, &internal))).await;
     /// });
     /// let (external_ref, internal_ref) = *fortified.borrow();
     /// assert_eq!(*external_ref, 1);
     /// assert_eq!(*internal_ref, 2);
     /// ```
-    pub fn new_async<'a, C, F>(cons: C) -> Self
+    pub fn new_async<C, F>(cons: C) -> Self
     where
-        T: WithLifetime<'a, Target = T>,
-        C: 'a + FnOnce(FortifyYielder<'a, T>) -> F,
+        C: 'a + FnOnce(FortifyYielder<T>) -> F,
         F: 'a + Future<Output = ()>,
     {
-        let mut context = FortifyYielderContext {
-            value: MaybeUninit::uninit(),
-            yielder_future_ptr: std::ptr::null_mut(),
-        };
-        let future = Box::into_raw(Box::new(cons(FortifyYielder {
-            context_ptr: &mut context,
-            marker: PhantomData,
-        })));
         let waker = nop_waker();
         let mut cx = Context::from_waker(&waker);
+        let mut data = FortifyYielderData {
+            value: MaybeUninit::uninit(),
+            tracker: FortifyYielderTracker {
+                cx_ptr: &cx as *const Context as *const (),
+                has_awaited: false,
+            }
+        };
+        let future = Box::into_raw(Box::new(cons(FortifyYielder(&mut data))));
         match Future::poll(unsafe { Pin::new_unchecked(&mut *future) }, &mut cx) {
             Poll::Ready(_) => {
                 unsafe { drop_box_from_raw::<F>(future as *mut ()) };
                 panic!("Future must await on FortifyYielder::yield_")
             }
             Poll::Pending => {
-                if !context.yielder_future_ptr.is_null() {
-                    unsafe {
-                        // Set `yielder_future_ptr` to null to prevent `context` from being
-                        // accessed after this call returns.
-                        (*context.yielder_future_ptr).0 = std::ptr::null_mut();
-                    }
+                if data.tracker.has_awaited {
                     Fortify {
-                        value: ManuallyDrop::new(unsafe { context.value.assume_init() }),
+                        value: ManuallyDrop::new(unsafe { data.value.assume_init() }),
                         data_raw: future as *mut (),
                         data_drop_fn: drop_box_from_raw::<F>,
                     }
@@ -174,43 +171,42 @@ impl<T> Fortify<T> {
             }
         }
     }
+}
 
-    /// Immutably borrows the value inside a [`Fortify`]. In practice, this can only be called when
-    /// `T` is covariant in its lifetime parameter, since the lifetime must be shortened to match
-    /// that of the `Fortify`. This is also the reason why there isn't a mutable version of this
-    /// function.
-    ///
-    /// For more general access to the wrapped value, see [`Fortify::with_ref`] and
-    /// [`Fortify::with_mut`].
-    pub fn borrow<'a>(&'a self) -> &'a T
-    where
-        T: WithLifetime<'a, Target = T>,
-    {
-        &self.value
+impl<'a, T: Lower<'a>> Fortify<T> {
+    /// Immutably borrows the value inside a [`Fortify`]. For more general access to the wrapped
+    /// value, see [`Fortify::with_ref`] and [`Fortify::with_mut`].
+    pub fn borrow(&'a self) -> &'a <T as Lower<'a>>::Target {
+        let value = &*self.value;
+        unsafe { transmute_copy(&value) }
     }
+}
 
+impl<T: for<'a> Lower<'a>> Fortify<T> {
     /// Executes a closure using an immutable reference to the value stored inside this [`Fortify`].
+    ///
+    /// Calls to `with_ref` can typically be replaced with and simplified using `borrow`. This
+    /// method is retained for consistency with `with_mut` and possible support for non-covariant
+    /// types (which can't use `borrow`) in the future.
     pub fn with_ref<'a, F, R>(&'a self, f: F) -> R
     where
-        T: for<'b> WithLifetime<'b>,
-        F: for<'b> FnOnce(&'a <T as WithLifetime<'b>>::Target) -> R,
+        F: for<'b> FnOnce(&'a <T as Lower<'b>>::Target) -> R,
     {
         let value = &*self.value;
-        f(unsafe { transmute(value) })
+        f(unsafe { transmute_copy(&value) })
     }
 
     /// Executes a closure using a mutable reference to the value stored inside this [`Fortify`].
     pub fn with_mut<'a, F, R>(&'a mut self, f: F) -> R
     where
-        T: for<'b> WithLifetime<'b>,
-        F: for<'b> FnOnce(&'a mut <T as WithLifetime<'b>>::Target) -> R,
+        F: for<'b> FnOnce(&'a mut <T as Lower<'b>>::Target) -> R,
     {
         let value = &mut *self.value;
-        f(unsafe { transmute(value) })
+        f(unsafe { transmute_copy(&value) })
     }
 }
 
-impl<'a, T> Fortify<&'a T> {
+impl<'a, T: Lower<'a, Target = T>> Fortify<&'a T> {
     /// Creates a [`Fortify`] by taking ownership of a [`Box`] and wrapping a reference to
     /// the value inside it.
     ///
@@ -222,12 +218,7 @@ impl<'a, T> Fortify<&'a T> {
     /// assert_eq!(**fortified.borrow(), 123);
     /// ```
     pub fn new_box_ref(value: Box<T>) -> Self {
-        let data_raw = Box::into_raw(value);
-        Self {
-            value: ManuallyDrop::new(unsafe { &*data_raw }),
-            data_raw: data_raw as *mut (),
-            data_drop_fn: drop_box_from_raw::<T>,
-        }
+        Fortify::new_box_dep(value, |inner| Lowered::new(&*inner))
     }
 }
 
@@ -244,22 +235,17 @@ impl<'a, T> Fortify<&'a mut T> {
     /// assert_eq!(**fortified.borrow(), 246);
     /// ```
     pub fn new_box_mut(value: Box<T>) -> Self {
-        let data_raw = Box::into_raw(value);
-        Self {
-            value: ManuallyDrop::new(unsafe { &mut *data_raw }),
-            data_raw: data_raw as *mut (),
-            data_drop_fn: drop_box_from_raw::<T>,
-        }
+        Fortify::new_box_dep(value, |inner| Lowered::new(inner))
     }
 }
 
-impl<T> From<T> for Fortify<T> {
+impl<'a, T: Lower<'a, Target = T> + 'a> From<T> for Fortify<T> {
     fn from(value: T) -> Self {
         Fortify::new(value)
     }
 }
 
-impl<'a, T> From<Box<T>> for Fortify<&'a T> {
+impl<'a, T: Lower<'a, Target = T>> From<Box<T>> for Fortify<&'a T> {
     fn from(value: Box<T>) -> Self {
         Fortify::new_box_ref(value)
     }
@@ -271,7 +257,7 @@ impl<'a, T> From<Box<T>> for Fortify<&'a mut T> {
     }
 }
 
-impl<T: Default> Default for Fortify<T> {
+impl<'a, T: Default + Lower<'a, Target = T> + 'a> Default for Fortify<T> {
     fn default() -> Self {
         Fortify::new(T::default())
     }
@@ -279,24 +265,31 @@ impl<T: Default> Default for Fortify<T> {
 
 impl<T> std::fmt::Debug for Fortify<T>
 where
-    for<'a> T: WithLifetime<'a>,
-    for<'a> <T as WithLifetime<'a>>::Target: std::fmt::Debug,
+    for<'a> T: Lower<'a>,
+    for<'a> <T as Lower<'a>>::Target: std::fmt::Debug,
 {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        self.with_ref(|inner| inner.fmt(f))
+        self.borrow().fmt(f)
     }
 }
 
 impl<T> std::fmt::Display for Fortify<T>
 where
-    for<'a> T: WithLifetime<'a>,
-    for<'a> <T as WithLifetime<'a>>::Target: std::fmt::Display,
+    for<'a> T: Lower<'a>,
+    for<'a> <T as Lower<'a>>::Target: std::fmt::Display,
 {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        self.with_ref(|inner| inner.fmt(f))
+        self.borrow().fmt(f)
     }
 }
 
+unsafe impl<'a, T: Lower<'a>> Lower<'a> for Fortify<T>
+where
+    <T as Lower<'a>>::Target: Sized,
+{
+    type Target = Fortify<<T as Lower<'a>>::Target>;
+}
+
 impl<T> Drop for Fortify<T> {
     fn drop(&mut self) {
         unsafe {
@@ -334,154 +327,51 @@ fn nop_waker() -> Waker {
     unsafe { Waker::from_raw(RawWaker::new(0 as *const (), VTABLE)) }
 }
 
-/// Wraps a value of type `T` and allows it to reference arbitrary supplementary data in the `'a`
-/// lifetime. This is isomorphic to `<T as WithLifetime<'a>>::Target`, but necessary to work around
-/// some [issues](https://stackoverflow.com/questions/60459609/how-do-i-return-an-associated-type-from-a-higher-ranked-trait-bound-trait)
-/// with type checking.
-pub struct FortifySource<'a, 'b, T> {
-    value: T,
-    marker: PhantomData<fn(&'a (), &'b ()) -> (&'a (), &'b ())>,
-}
-
-impl<'a, 'b, T> FortifySource<'a, 'b, T> {
-    /// Constructs a [`FortifySource`] from its wrapped value.
-    pub fn new<O>(value: O) -> Self
-    where
-        T: WithLifetime<'a, Target = O>,
-        O: WithLifetime<'b, Target = T>,
-    {
-        let value = unsafe { transmute_copy(&value) };
-        FortifySource {
-            value,
-            marker: PhantomData,
-        }
-    }
-
-    /// Unpacks this [`FortifySource`] wrapper.
-    pub fn into_inner(self) -> <T as WithLifetime<'a>>::Target
-    where
-        T: WithLifetime<'a>,
-    {
-        let res = unsafe { transmute_copy(&self.value) };
-        forget(self.value);
-        res
-    }
-}
-
 /// A helper interface used by the [`Fortify::new_async`] constructor.
-pub struct FortifyYielder<'a, T> {
-    context_ptr: *mut FortifyYielderContext<T>,
-    marker: PhantomData<&'a ()>,
-}
+pub struct FortifyYielder<T>(*mut FortifyYielderData<T>);
 
-impl<'a, T> FortifyYielder<'a, T> {
+impl<T> FortifyYielder<T> {
     /// Provides the [`Fortify`] value to this [`FortifyYielder`] and returns a [`Future`] that may
     /// be awaited to suspend execution.
-    pub fn yield_<'b, O>(self, value: O) -> impl Future<Output = ()> + 'b
-    where
-        T: WithLifetime<'b, Target = O>,
-        O: WithLifetime<'a, Target = T>,
-    {
+    pub fn yield_<'a>(self, value: Lowered<'a, T>) -> impl Future<Output = ()> + 'a {
         unsafe {
-            let target = &mut *(self.context_ptr as *mut FortifyYielderContext<O>);
-            target.value.write(value);
-            FortifyYielderFuture(&mut target.yielder_future_ptr)
+            let target = &mut *self.0;
+            target.value.write(value.value);
+            FortifyYielderFuture(&mut target.tracker)
         }
     }
 }
 
-struct FortifyYielderContext<T> {
+struct FortifyYielderData<T> {
     value: MaybeUninit<T>,
-    yielder_future_ptr: *mut FortifyYielderFuture,
+    tracker: FortifyYielderTracker,
 }
 
-struct FortifyYielderFuture(*mut *mut FortifyYielderFuture);
+struct FortifyYielderTracker {
+    cx_ptr: *const (),
+    has_awaited: bool,
+}
+
+struct FortifyYielderFuture(*mut FortifyYielderTracker);
 
 impl Future for FortifyYielderFuture {
     type Output = ();
 
-    fn poll(self: Pin<&mut Self>, _: &mut Context<'_>) -> Poll<Self::Output> {
-        if !self.0.is_null() {
-            unsafe {
-                // Inform the context that the yield future has been created and pinned. This is a
-                // necessary precondition to using the yielded value.
-                let self_mut = Pin::into_inner_unchecked(self);
-                *self_mut.0 = &mut *self_mut;
+    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
+        unsafe {
+            // Verify the context to ensure that this future is being polled by `new_async` rather
+            // than by the user.
+            let tracker = &mut *self.as_ref().0;
+            if tracker.cx_ptr == (cx as *const Context as *const ()) {
+                // Inform `new_async` that the future has been awaited. This enables the value
+                // written to `FortifyYielderData` to be used.
+                tracker.has_awaited = true;
             }
         }
         Poll::Pending
     }
 }
 
-impl Drop for FortifyYielderFuture {
-    fn drop(&mut self) {
-        if !self.0.is_null() {
-            unsafe {
-                // Inform the context that the yielder future has been dropped. This indicates
-                // that it is no longer safe to use the yielded value.
-                *self.0 = std::ptr::null_mut();
-            }
-        }
-    }
-}
-
-/// Indicates that a type has a "primary" lifetime parameter and that parameter can be
-/// replaced with another lifetime (`'a`) to produce the type `WithLifetime<'a>::Target`. This
-/// trait can be automatically derived. When doing so, the first lifetime parameter in the
-/// parameters list will be treated as the "primary" lifetime parameter.
-pub trait WithLifetime<'a> {
-    /// The type resulting from replacing the "primary" lifetime of `Self` with `'a`.
-    type Target: WithLifetime<'a, Target = Self::Target>;
-}
-
-impl<'a, 'b, T: 'b + ?Sized> WithLifetime<'b> for &'a T {
-    type Target = &'b T;
-}
-
-impl<'a, 'b, T: 'b + ?Sized> WithLifetime<'b> for &'a mut T {
-    type Target = &'b mut T;
-}
-
-impl<'a, T: WithLifetime<'a> + ?Sized> WithLifetime<'a> for Box<T> {
-    type Target = Box<<T as WithLifetime<'a>>::Target>;
-}
-
-impl<'a, T: WithLifetime<'a>, const N: usize> WithLifetime<'a> for [T; N] {
-    type Target = [<T as WithLifetime<'a>>::Target; N];
-}
-
-impl<'a, T: WithLifetime<'a>> WithLifetime<'a> for Fortify<T> {
-    type Target = Fortify<<T as WithLifetime<'a>>::Target>;
-}
-
-impl<'a, A: WithLifetime<'a>, B: WithLifetime<'a>> WithLifetime<'a> for (A, B) {
-    type Target = (
-        <A as WithLifetime<'a>>::Target,
-        <B as WithLifetime<'a>>::Target,
-    );
-}
-
-impl<'a, A: WithLifetime<'a>, B: WithLifetime<'a>, C: WithLifetime<'a>> WithLifetime<'a>
-    for (A, B, C)
-{
-    type Target = (
-        <A as WithLifetime<'a>>::Target,
-        <B as WithLifetime<'a>>::Target,
-        <C as WithLifetime<'a>>::Target,
-    );
-}
-
-impl<'a, A: WithLifetime<'a>, B: WithLifetime<'a>, C: WithLifetime<'a>, D: WithLifetime<'a>>
-    WithLifetime<'a> for (A, B, C, D)
-{
-    type Target = (
-        <A as WithLifetime<'a>>::Target,
-        <B as WithLifetime<'a>>::Target,
-        <C as WithLifetime<'a>>::Target,
-        <D as WithLifetime<'a>>::Target,
-    );
-}
-
 /// A helper macro for creating a `Fortify` using generator-like syntax. The macro takes a block of
 /// statements that ends with a `yield` of some expression. The block will be executed up to the
 /// `yield` statement, at which point the value of expression will be bundled with the suspended
@@ -503,13 +393,13 @@ impl<'a, A: WithLifetime<'a>, B: WithLifetime<'a>, C: WithLifetime<'a>, D: WithL
 #[macro_export]
 macro_rules! fortify {
     (@INNER $y:ident , yield $res:expr ;) => {
-        $y.yield_($res).await
+        $y.yield_(Lowered::new($res)).await
     };
     (@INNER $y:ident , $st:stmt ; $($t:tt)*) => {
         { $st fortify!(@INNER $y , $($t)*) }
     };
     ($($t:tt)*) => {
-        $crate::Fortify::new_async(|y| async {
+        $crate::Fortify::new_async(move |y| async move {
             fortify!(@INNER y , $($t)*)
         })
     };
